"""
================================================================================

  This file is part of the AMUN source code, a program to perform
  Newtonian or relativistic magnetohydrodynamical simulations on uniform or
  adaptive mesh.

  Copyright (C) 2018-2021 Grzegorz Kowal <grzegorz@amuncode.org>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

================================================================================

 module: AMUN

  Python module with subroutines to read AMUN code HDF5 files.

  The only requirements for this package are:

     - h5py
     - numpy
     - scipy (optional, for data interpolation)
     - xxhash (optional, for data consistency verification)
     - zstd (optional, for data compression)

--------------------------------------------------------------------------------
"""
from .interpolation import interpolate
import numpy as np
import os.path as op
import sys
try:
    from xxhash import *
    xxhash_available = True
except ImportError:
    xxhash_available = False
try:
    import zstandard as zstd
    zstd_available = True
except ImportError:
    zstd_available = False
try:
    import lz4.frame as lz4
    lz4_available = True
except ImportError:
    lz4_available = False
try:
    import lzma
    lzma_available = True
except ImportError:
    lzma_available = False


class AmunXML:
    """AMUN XML snapshot class"""

    def __init__(self, path):

        import xml.etree.ElementTree as ET

        if op.isdir(path):
            mfile = op.join(path, 'metadata.xml')
            if op.exists(mfile):
                self.path = path
                self.metafile = mfile
                self.format = "xml"
            else:
                print("{} does not contain 'metadata.xml'!".format(path))
                return None
        else:
            mfile = op.basename(path)
            if mfile == 'metadata.xml':
                self.path = op.dirname(path)
                self.metafile = path
                self.format = "xml"
            else:
                print("{} is not 'metadata.xml'!".format(path))
                return None

        tree = ET.parse(self.metafile)
        root = tree.getroot()
        if root.tag == 'AMUNFile':
            self.attributes = dict()
            self.binaries = dict()
            for child in root:
                if child.tag == 'BinaryFiles':
                    for item in child:
                        self.binaries[item.attrib['name']] = item.attrib
                        self.binaries[item.attrib['name']]['file'] = item.text
                else:
                    for item in child:
                        if item.tag == 'Attribute':
                            if item.attrib['type'] == 'double':
                                self.attributes[item.attrib['name']] = float(item.text)
                            elif item.attrib['type'] == 'integer':
                                self.attributes[item.attrib['name']] = int(item.text)
                            else:
                                self.attributes[item.attrib['name']] = item.text
        else:
            print("{} is not an AMUN snapshot!".format(self.metafile))
            return None

        self.chunks = dict()
        for n in range(self.attributes['nprocs']):
            mfile = op.join(self.path, 'datablocks_{:06d}.xml'.format(n))
            tree = ET.parse(mfile)
            root = tree.getroot()
            self.chunks[n] = dict()
            for item in root.findall('./BinaryFiles/Attribute'):
                self.chunks[n][item.attrib['name']] = item.attrib
                self.chunks[n][item.attrib['name']]['file'] = item.text
            for item in root.findall('./DataBlocks/Attribute'):
                if item.attrib['name'] == 'dblocks':
                    self.chunks[n]['dblocks'] = int(item.text)

        if 'variables' in self.attributes:
            variables = []
            for var in self.attributes['variables'].split():
                variables.append(var)

            variables.append('level')
            if 'dens' in variables:
                variables.append('logd')
            if 'pres' in variables:
                variables.append('logp')
            if all(v in variables for v in ['velx','vely','velz']):
                variables.append('velocity_vector')
                variables.append('velo')
                variables.append('divv')
                variables.append('vorx')
                variables.append('vory')
                variables.append('vorz')
                variables.append('vort')
                variables.append('dvxdx')
                variables.append('dvxdy')
                variables.append('dvxdz')
                variables.append('dvydx')
                variables.append('dvydy')
                variables.append('dvydz')
                variables.append('dvzdx')
                variables.append('dvzdy')
                variables.append('dvzdz')
            if all(v in variables for v in ['magx','magy','magz']):
                variables.append('magnetic_vector')
                variables.append('magn')
                variables.append('divb')
                variables.append('curx')
                variables.append('cury')
                variables.append('curz')
                variables.append('curr')
                variables.append('dbxdx')
                variables.append('dbxdy')
                variables.append('dbxdz')
                variables.append('dbydx')
                variables.append('dbydy')
                variables.append('dbydz')
                variables.append('dbzdx')
                variables.append('dbzdy')
                variables.append('dbzdz')
            if 'pres' in variables and 'adiabatic_index' in self.attributes:
                variables.append('eint')
            if all(v in variables for v in ['dens','pres']):
                variables.append('temp')
            if self.attributes['eqsys'] in [ 'hd', 'mhd' ] \
                            and all(v in variables for v in ['dens','velo']):
                variables.append('ekin')
            if self.attributes['eqsys'] in [ 'mhd', 'srmhd' ] \
                            and 'magn' in variables:
                variables.append('emag')
            if 'velo' in variables and 'magn' in variables:
                variables.append('elex')
                variables.append('eley')
                variables.append('elez')
                variables.append('eele')
            if all(v in variables for v in ['ekin','eint']):
                variables.append('etot')
            if self.attributes['eqsys'] in [ 'srhd', 'srmhd' ] \
                            and 'velo' in variables:
                variables.append('lore')
        else:
            print("Cannot determine available variables from {}!".format(self.metafile))
            return None

        self.attributes['xlen'] = self.attributes['xmax'] - self.attributes['xmin']
        self.attributes['ylen'] = self.attributes['ymax'] - self.attributes['ymin']
        if self.attributes['ndims'] == 3:
            self.attributes['zlen'] = self.attributes['zmax'] - self.attributes['zmin']

        self.cell_size = dict()
        for l in range(self.attributes['maxlev']):
            n = self.attributes['xblocks'] * self.attributes['ncells'] * 2**l
            h = self.attributes['xlen'] / n
            self.cell_size[l+1] = h

        self.variables = variables

        # prepare meta block levels and coordinates
        mset = self.__read_binary_meta('fields')

        nf    = int(mset.size / self.attributes['nleafs'])
        mset  = np.reshape(mset, [nf,self.attributes['nleafs']])

        self.index = dict()
        self.level = dict()
        self.coord = dict()
        for n in range(self.attributes['nleafs']):
            self.index[mset[0,n]] = n
            self.level[mset[0,n]] = mset[  1,n]
            self.coord[mset[0,n]] = mset[2:5,n]
        self.ml = mset[1,:].max()

        self.bounds = np.reshape(self.__read_binary_meta('bounds', dtype='float64'), \
                                [2, self.attributes['ndims'], self.attributes['nleafs']])


    def __check_digest(self, dfile, data, digest):
        if xxhash_available:
            if digest.lower() != xxh64(data).hexdigest():
                print("File '{}' seems to be corrupted!".format(dfile))


    def __read_binary_meta(self, aname, dtype='int32'):

        dfile = op.join(self.path, self.binaries[aname]['file'])
        if op.exists(dfile):
            if 'compression_format' in self.binaries[aname]:
                cformat = self.binaries[aname]['compression_format']
                if cformat == 'zstd':
                    if zstd_available:
                        dctx = zstd.ZstdDecompressor()
                        with open(dfile, mode ='rb') as f:
                            compressed = f.read()
                        if 'compressed_digest' in self.binaries[aname]:
                            self.__check_digest(dfile, compressed, self.binaries[aname]['compressed_digest'])
                        data = dctx.decompress(compressed)
                        if 'digest' in self.binaries[aname]:
                            self.__check_digest(dfile, data, self.binaries[aname]['digest'])
                        return np.frombuffer(data, dtype=dtype)
                    else:
                        print('ZSTD module unavailable. Please, install Zstandard for Python.')
                        return None
                elif cformat == 'lz4':
                    if lz4_available:
                        with open(dfile, mode ='rb') as f:
                            compressed = f.read()
                        if 'compressed_digest' in self.binaries[aname]:
                            self.__check_digest(dfile, compressed, self.binaries[aname]['compressed_digest'])
                        data = lz4.decompress(compressed)
                        if 'digest' in self.binaries[aname]:
                            self.__check_digest(dfile, data, self.binaries[aname]['digest'])
                        return np.frombuffer(data, dtype=dtype)
                    else:
                        print('LZ4 module unavailable. Please, install LZ4 for Python.')
                        return None
                elif cformat == 'lzma':
                    if lzma_available:
                        with open(dfile, mode ='rb') as f:
                            compressed = f.read()
                        if 'compressed_digest' in self.binaries[aname]:
                            self.__check_digest(dfile, compressed, self.binaries[aname]['compressed_digest'])
                        data = lzma.decompress(compressed)
                        if 'digest' in self.binaries[aname]:
                            self.__check_digest(dfile, data, self.binaries[aname]['digest'])
                        return np.frombuffer(data, dtype=dtype)
                    else:
                        print('LZMA module unavailable. Please, install LZMA for Python.')
                        return None
                else:
                    print('Compression format {} unavailable. Please, install corresponding Python module'.format(cformat))
                    return None
            else:
                with open(dfile, mode ='rb') as f:
                    data = f.read()
                if 'digest' in self.binaries[aname]:
                    self.__check_digest(dfile, data, self.binaries[aname]['digest'])
                return np.frombuffer(data, dtype=dtype)
        else:
            return None


    def __read_binary_data(self, n, var, dtype='float64'):

        dfile = op.join(self.path, self.chunks[n][var]['file'])
        if op.exists(dfile):
            if 'compression_format' in self.chunks[n][var]:
                cformat = self.chunks[n][var]['compression_format']
                if cformat == 'zstd':
                    if zstd_available:
                        dctx = zstd.ZstdDecompressor()
                        with open(dfile, mode ='rb') as f:
                            compressed = f.read()
                        if 'compressed_digest' in self.chunks[n][var]:
                            self.__check_digest(dfile, compressed, self.chunks[n][var]['compressed_digest'])
                        data = dctx.decompress(compressed)
                        if 'digest' in self.chunks[n][var]:
                            self.__check_digest(dfile, data, self.chunks[n][var]['digest'])
                        return np.frombuffer(data, dtype=dtype)
                    else:
                        print('ZSTD module unavailable. Please, install Zstandard for Python.')
                        return None
                elif cformat == 'lz4':
                    if lz4_available:
                        with open(dfile, mode ='rb') as f:
                            compressed = f.read()
                        if 'compressed_digest' in self.chunks[n][var]:
                            self.__check_digest(dfile, compressed, self.chunks[n][var]['compressed_digest'])
                        data = lz4.decompress(compressed)
                        if 'digest' in self.chunks[n][var]:
                            self.__check_digest(dfile, data, self.chunks[n][var]['digest'])
                        return np.frombuffer(data, dtype=dtype)
                    else:
                        print('LZ4 module unavailable. Please, install LZ4 for Python.')
                        return None
                elif cformat == 'lzma':
                    if lzma_available:
                        with open(dfile, mode ='rb') as f:
                            compressed = f.read()
                        if 'compressed_digest' in self.chunks[n][var]:
                            self.__check_digest(dfile, compressed, self.chunks[n][var]['compressed_digest'])
                        data = lzma.decompress(compressed)
                        if 'digest' in self.chunks[n][var]:
                            self.__check_digest(dfile, data, self.chunks[n][var]['digest'])
                        return np.frombuffer(data, dtype=dtype)
                    else:
                        print('LZMA module unavailable. Please, install LZMA for Python.')
                        return None
                else:
                    print('Compression format {} unavailable. Please, install corresponding Python module'.format(cformat))
                    return None
            else:
                with open(dfile, mode ='rb') as f:
                    data = f.read()
                if 'digest' in self.chunks[n][var]:
                    self.__check_digest(dfile, data, self.chunks[n][var]['digest'])
                return np.frombuffer(data, dtype=dtype)
        else:
            return None


    def __get_dataset(self, n, var, cm):

        ids = self.__read_binary_data(n, 'ids', dtype='int32')

        if var == 'level':
            dset = np.zeros(cm)
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] = self.level[ids[p]]
        elif var == 'logd':
            dset = np.log10(self.__read_binary_data(n, 'dens'))
            dset = np.reshape(dset, cm)
        elif var == 'logp':
            dset = np.log10(self.__read_binary_data(n, 'pres'))
            dset = np.reshape(dset, cm)
        elif var == 'velo':
            tmp = self.__read_binary_data(n, 'velx')
            dset = tmp**2
            tmp = self.__read_binary_data(n, 'vely')
            dset += tmp**2
            tmp = self.__read_binary_data(n, 'velz')
            dset += tmp**2
            dset = np.reshape(np.sqrt(dset), cm)
        elif var == 'velocity_vector':
            dset = [np.reshape(self.__read_binary_data(n, 'velx'), cm), \
                    np.reshape(self.__read_binary_data(n, 'vely'), cm), \
                    np.reshape(self.__read_binary_data(n, 'velz'), cm)]
        elif var == 'magn':
            tmp = self.__read_binary_data(n, 'magx')
            dset = tmp**2
            tmp = self.__read_binary_data(n, 'magy')
            dset += tmp**2
            tmp = self.__read_binary_data(n, 'magz')
            dset += tmp**2
            dset = np.reshape(np.sqrt(dset), cm)
        elif var == 'magnetic_vector':
            dset = [np.reshape(self.__read_binary_data(n, 'magx'), cm), \
                    np.reshape(self.__read_binary_data(n, 'magy'), cm), \
                    np.reshape(self.__read_binary_data(n, 'magz'), cm)]
        elif var == 'ekin':
            tmp = self.__read_binary_data(n, 'velx')
            dset = tmp**2
            tmp = self.__read_binary_data(n, 'vely')
            dset += tmp**2
            tmp = self.__read_binary_data(n, 'velz')
            dset += tmp**2
            tmp = self.__read_binary_data(n, 'dens')
            dset *= tmp
            dset *= 0.5
            dset = np.reshape(dset, cm)
        elif var == 'emag':
            tmp = self.__read_binary_data(n, 'magx')
            dset = tmp**2
            tmp = self.__read_binary_data(n, 'magy')
            dset += tmp**2
            tmp = self.__read_binary_data(n, 'magz')
            dset += tmp**2
            dset *= 0.5
            dset = np.reshape(dset, cm)
        elif var == 'elex':
            by = np.reshape(self.__read_binary_data(n, 'magy'), cm)
            bz = np.reshape(self.__read_binary_data(n, 'magz'), cm)
            vy = np.reshape(self.__read_binary_data(n, 'vely'), cm)
            vz = np.reshape(self.__read_binary_data(n, 'velz'), cm)
            dset = vy * bz - vz * by
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (np.roll(by,  1, axis=0) - np.roll(by, -1, axis=0))
                    tmp += (np.roll(bz, -1, axis=1) - np.roll(bz,  1, axis=1))
                else:
                    tmp  = (np.roll(bz, -1, axis=0) - np.roll(bz,  1, axis=0))

                for p in range(self.chunks[n]['dblocks']):
                    tmp[...,p] /= self.cell_size[self.level[ids[p]]]
                dset -= 0.5 * self.attributes['resistivity'] * tmp
        elif var == 'eley':
            bx = np.reshape(self.__read_binary_data(n, 'magx'), cm)
            bz = np.reshape(self.__read_binary_data(n, 'magz'), cm)
            vx = np.reshape(self.__read_binary_data(n, 'velx'), cm)
            vz = np.reshape(self.__read_binary_data(n, 'velz'), cm)
            dset = vz * bx - vx * bz
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (np.roll(bx, -1, axis=0) - np.roll(bx,  1, axis=0))
                    tmp += (np.roll(bz,  1, axis=2) - np.roll(bz, -1, axis=2))
                else:
                    tmp  = (np.roll(bz,  1, axis=1) - np.roll(bz, -1, axis=1))

                for p in range(self.chunks[n]['dblocks']):
                    tmp[...,p] /= self.cell_size[self.level[ids[p]]]
                dset -= 0.5 * self.attributes['resistivity'] * tmp
        elif var == 'elez':
            bx = np.reshape(self.__read_binary_data(n, 'magx'), cm)
            by = np.reshape(self.__read_binary_data(n, 'magy'), cm)
            vx = np.reshape(self.__read_binary_data(n, 'velx'), cm)
            vy = np.reshape(self.__read_binary_data(n, 'vely'), cm)
            dset = vx * by - vy * bx
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (np.roll(bx,  1, axis=1) - np.roll(bx, -1, axis=1))
                    tmp += (np.roll(by, -1, axis=2) - np.roll(by,  1, axis=2))
                else:
                    tmp  = (np.roll(bx,  1, axis=0) - np.roll(bx, -1, axis=0))
                    tmp += (np.roll(by, -1, axis=1) - np.roll(by,  1, axis=1))

                for p in range(self.chunks[n]['dblocks']):
                    tmp[...,p] /= self.cell_size[self.level[ids[p]]]
                dset -= 0.5 * self.attributes['resistivity'] * tmp
        elif var == 'eele':
            b1 = np.reshape(self.__read_binary_data(n, 'magy'), cm)
            b2 = np.reshape(self.__read_binary_data(n, 'magz'), cm)
            v1 = np.reshape(self.__read_binary_data(n, 'vely'), cm)
            v2 = np.reshape(self.__read_binary_data(n, 'velz'), cm)
            dtmp = v1 * b2 - v2 * b1
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (np.roll(b1,  1, axis=0) - np.roll(b1, -1, axis=0))
                    tmp += (np.roll(b2, -1, axis=1) - np.roll(b2,  1, axis=1))
                else:
                    tmp  = (np.roll(b2, -1, axis=0) - np.roll(b2,  1, axis=0))

                for p in range(self.chunks[n]['dblocks']):
                    tmp[...,p] /= self.cell_size[self.level[ids[p]]]
                dtmp -= 0.5 * self.attributes['resistivity'] * tmp
            dset  = dtmp**2

            b1 = np.reshape(self.__read_binary_data(n, 'magx'), cm)
            v1 = np.reshape(self.__read_binary_data(n, 'velx'), cm)
            dtmp = v2 * b1 - v1 * b2
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (np.roll(b1, -1, axis=0) - np.roll(b1,  1, axis=0))
                    tmp += (np.roll(b2,  1, axis=2) - np.roll(b2, -1, axis=2))
                else:
                    tmp  = (np.roll(b2,  1, axis=1) - np.roll(b2, -1, axis=1))

                for p in range(self.chunks[n]['dblocks']):
                    tmp[...,p] /= self.cell_size[self.level[ids[p]]]
                dtmp -= 0.5 * self.attributes['resistivity'] * tmp
            dset += dtmp**2

            b2 = np.reshape(self.__read_binary_data(n, 'magy'), cm)
            v2 = np.reshape(self.__read_binary_data(n, 'vely'), cm)
            dtmp = v1 * b2 - v2 * b1
            if self.attributes['resistivity'] > 0:
                if self.attributes['ndims'] == 3:
                    tmp  = (np.roll(b1,  1, axis=1) - np.roll(b1, -1, axis=1))
                    tmp += (np.roll(b2, -1, axis=2) - np.roll(b2,  1, axis=2))
                else:
                    tmp  = (np.roll(b1,  1, axis=0) - np.roll(b1, -1, axis=0))
                    tmp += (np.roll(b2, -1, axis=1) - np.roll(b2,  1, axis=1))

                for p in range(self.chunks[n]['dblocks']):
                    tmp[...,p] /= self.cell_size[self.level[ids[p]]]
                dtmp -= 0.5 * self.attributes['resistivity'] * tmp
            dset += dtmp**2
            dset = np.sqrt(dset)
        elif var == 'eint':
            dset = self.__read_binary_data(n, 'pres')
            dset *= 1.0 / (self.attributes('adiabatic_index') - 1)
            dset = np.reshape(dset, cm)
        elif var == 'temp':
            dset = self.__read_binary_data(n, 'pres') / self.__read_binary_data(n, 'dens')
            dset = np.reshape(dset, cm)
        elif var == 'etot':
            tmp = self.__read_binary_data(n, 'velx')
            dset = tmp**2
            tmp = self.__read_binary_data(n, 'vely')
            dset += tmp**2
            tmp = self.__read_binary_data(n, 'velz')
            dset += tmp**2
            tmp = self.__read_binary_data(n, 'dens')
            dset *= tmp
            tmp = self.__read_binary_data(n, 'pres')
            dset += 2.0 / (self.attributes('adiabatic_index') - 1) * tmp
            if 'magn' in self.variables:
                tmp = self.__read_binary_data(n, 'magx')
                dset = tmp**2
                tmp = self.__read_binary_data(n, 'magy')
                dset += tmp**2
                tmp = self.__read_binary_data(n, 'magz')
                dset += tmp**2
            dset *= 0.5
            dset = np.reshape(dset, cm)
        elif var == 'lorentz':
            tmp = self.__read_binary_data(n, 'velx')
            dset = tmp**2
            tmp = self.__read_binary_data(n, 'vely')
            dset += tmp**2
            tmp = self.__read_binary_data(n, 'velz')
            dset += tmp**2
            dset = 1.0 / np.sqrt(1.0 - dset)
            dset = np.reshape(dset, cm)
        elif var == 'divv':
            p = self.attributes['ndims'] - 1
            tmp = np.reshape(self.__read_binary_data(n, 'velx'), cm)
            dset  = (np.roll(tmp, -1, axis=p) \
                   - np.roll(tmp,  1, axis=p))
            p -= 1
            tmp = np.reshape(self.__read_binary_data(n, 'vely'), cm)
            dset += (np.roll(tmp, -1, axis=p) \
                   - np.roll(tmp,  1, axis=p))
            p -= 1
            if p >= 0:
                tmp = np.reshape(self.__read_binary_data(n, 'velz'), cm)
                dset += (np.roll(tmp, -1, axis=p) \
                       - np.roll(tmp,  0, axis=p))

            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'vorx':
            if self.attributes['ndims'] == 3:
                tmp = np.reshape(self.__read_binary_data(n, 'vely'), cm)
                dset  = (np.roll(tmp,  1, axis=0) \
                       - np.roll(tmp, -1, axis=0))
                tmp = np.reshape(self.__read_binary_data(n, 'velz'), cm)
                dset += (np.roll(tmp, -1, axis=1) \
                       - np.roll(tmp,  1, axis=1))
            else:
                tmp = np.reshape(self.__read_binary_data(n, 'velz'), cm)
                dset  = (np.roll(tmp, -1, axis=0) \
                       - np.roll(tmp,  1, axis=0))

            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'vory':
            if self.attributes['ndims'] == 3:
                tmp = np.reshape(self.__read_binary_data(n, 'velx'), cm)
                dset  = (np.roll(tmp, -1, axis=0) \
                       - np.roll(tmp,  1, axis=0))
                tmp = np.reshape(self.__read_binary_data(n, 'velz'), cm)
                dset += (np.roll(tmp,  1, axis=2) \
                       - np.roll(tmp, -1, axis=2))
            else:
                tmp = np.reshape(self.__read_binary_data(n, 'velz'), cm)
                dset  = (np.roll(tmp,  1, axis=1) \
                       - np.roll(tmp, -1, axis=1))

            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'vorz':
            if self.attributes['ndims'] == 3:
                tmp = np.reshape(self.__read_binary_data(n, 'velx'), cm)
                dset  = (np.roll(tmp,  1, axis=1) \
                       - np.roll(tmp, -1, axis=1))
                tmp = np.reshape(self.__read_binary_data(n, 'vely'), cm)
                dset += (np.roll(tmp, -1, axis=2) \
                       - np.roll(tmp,  1, axis=2))
            else:
                tmp = np.reshape(self.__read_binary_data(n, 'velx'), cm)
                dset  = (np.roll(tmp,  1, axis=0) \
                       - np.roll(tmp, -1, axis=0))
                tmp = np.reshape(self.__read_binary_data(n, 'vely'), cm)
                dset += (np.roll(tmp, -1, axis=1) \
                       - np.roll(tmp,  1, axis=1))

            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'vort':
            if self.attributes['ndims'] == 3:
                tmp = np.reshape(self.__read_binary_data(n, 'velx'), cm)
                wy  = (np.roll(tmp, -1, axis=0) \
                     - np.roll(tmp,  1, axis=0))
                wz  = (np.roll(tmp,  1, axis=1) \
                     - np.roll(tmp, -1, axis=1))
                tmp = np.reshape(self.__read_binary_data(n, 'vely'), cm)
                wz += (np.roll(tmp, -1, axis=2) \
                     - np.roll(tmp,  1, axis=2))
                wx  = (np.roll(tmp,  1, axis=0) \
                     - np.roll(tmp, -1, axis=0))
                tmp = np.reshape(self.__read_binary_data(n, 'velz'), cm)
                wx += (np.roll(tmp, -1, axis=1) \
                     - np.roll(tmp,  1, axis=1))
                wy += (np.roll(tmp,  1, axis=2) \
                     - np.roll(tmp, -1, axis=2))
            else:
                tmp = np.reshape(self.__read_binary_data(n, 'velx'), cm)
                wz  = (np.roll(tmp,  1, axis=0) \
                     - np.roll(tmp, -1, axis=0))
                tmp = np.reshape(self.__read_binary_data(n, 'vely'), cm)
                wz += (np.roll(tmp, -1, axis=1) \
                     - np.roll(tmp,  1, axis=1))
                tmp = np.reshape(self.__read_binary_data(n, 'velz'), cm)
                wx  = (np.roll(tmp, -1, axis=0) \
                     - np.roll(tmp,  1, axis=0))
                wy  = (np.roll(tmp,  1, axis=1) \
                     - np.roll(tmp, -1, axis=1))

            dset = 0.5 * np.sqrt(wx**2 + wy**2 + wz**2)
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'dvxdx':
            p = self.attributes['ndims'] - 1
            tmp = np.reshape(self.__read_binary_data(n, 'velx'), cm)
            dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'dvxdy':
            p = self.attributes['ndims'] - 2
            tmp = np.reshape(self.__read_binary_data(n, 'velx'), cm)
            dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'dvxdz':
            p = self.attributes['ndims'] - 3
            tmp = np.reshape(self.__read_binary_data(n, 'velx'), cm)
            if p >= 0:
                dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
                dset *= 0.5
                for p in range(self.chunks[n]['dblocks']):
                    dset[...,p] /= self.cell_size[self.level[ids[p]]]
            else:
                dset = np.zeros_like(tmp)
        elif var == 'dvydx':
            p = self.attributes['ndims'] - 1
            tmp = np.reshape(self.__read_binary_data(n, 'vely'), cm)
            dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'dvydy':
            p = self.attributes['ndims'] - 2
            tmp = np.reshape(self.__read_binary_data(n, 'vely'), cm)
            dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'dvydz':
            p = self.attributes['ndims'] - 3
            tmp = np.reshape(self.__read_binary_data(n, 'vely'), cm)
            if p >= 0:
                dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
                dset *= 0.5
                for p in range(self.chunks[n]['dblocks']):
                    dset[...,p] /= self.cell_size[self.level[ids[p]]]
            else:
                dset = np.zeros_like(tmp)
        elif var == 'dvzdx':
            p = self.attributes['ndims'] - 1
            tmp = np.reshape(self.__read_binary_data(n, 'velz'), cm)
            dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'dvzdy':
            p = self.attributes['ndims'] - 2
            tmp = np.reshape(self.__read_binary_data(n, 'velz'), cm)
            dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'dvzdz':
            p = self.attributes['ndims'] - 3
            tmp = np.reshape(self.__read_binary_data(n, 'velz'), cm)
            if p >= 0:
                dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
                dset *= 0.5
                for p in range(self.chunks[n]['dblocks']):
                    dset[...,p] /= self.cell_size[self.level[ids[p]]]
            else:
                dset = np.zeros_like(tmp)
        elif var == 'divb':
            p = self.attributes['ndims'] - 1
            tmp = np.reshape(self.__read_binary_data(n, 'magx'), cm)
            dset  = (np.roll(tmp, -1, axis=p) \
                   - np.roll(tmp,  1, axis=p))
            p -= 1
            tmp = np.reshape(self.__read_binary_data(n, 'magy'), cm)
            dset += (np.roll(tmp, -1, axis=p) \
                   - np.roll(tmp,  1, axis=p))
            p -= 1
            if p >= 0:
                tmp = np.reshape(self.__read_binary_data(n, 'magz'), cm)
                dset += (np.roll(tmp, -1, axis=p) \
                       - np.roll(tmp,  0, axis=p))

            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'curx':
            if self.attributes['ndims'] == 3:
                tmp = np.reshape(self.__read_binary_data(n, 'magy'), cm)
                dset  = (np.roll(tmp,  1, axis=0) \
                       - np.roll(tmp, -1, axis=0))
                tmp = np.reshape(self.__read_binary_data(n, 'magz'), cm)
                dset += (np.roll(tmp, -1, axis=1) \
                       - np.roll(tmp,  1, axis=1))
            else:
                tmp = np.reshape(self.__read_binary_data(n, 'magz'), cm)
                dset  = (np.roll(tmp, -1, axis=0) \
                       - np.roll(tmp,  1, axis=0))

            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'cury':
            if self.attributes['ndims'] == 3:
                tmp = np.reshape(self.__read_binary_data(n, 'magx'), cm)
                dset  = (np.roll(tmp, -1, axis=0) \
                       - np.roll(tmp,  1, axis=0))
                tmp = np.reshape(self.__read_binary_data(n, 'magz'), cm)
                dset += (np.roll(tmp,  1, axis=2) \
                       - np.roll(tmp, -1, axis=2))
            else:
                tmp = np.reshape(self.__read_binary_data(n, 'magz'), cm)
                dset  = (np.roll(tmp,  1, axis=1) \
                       - np.roll(tmp, -1, axis=1))

            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'curz':
            if self.attributes['ndims'] == 3:
                tmp = np.reshape(self.__read_binary_data(n, 'magx'), cm)
                dset  = (np.roll(tmp,  1, axis=1) \
                       - np.roll(tmp, -1, axis=1))
                tmp = np.reshape(self.__read_binary_data(n, 'magy'), cm)
                dset += (np.roll(tmp, -1, axis=2) \
                       - np.roll(tmp,  1, axis=2))
            else:
                tmp = np.reshape(self.__read_binary_data(n, 'magx'), cm)
                dset  = (np.roll(tmp,  1, axis=0) \
                       - np.roll(tmp, -1, axis=0))
                tmp = np.reshape(self.__read_binary_data(n, 'magy'), cm)
                dset += (np.roll(tmp, -1, axis=1) \
                       - np.roll(tmp,  1, axis=1))

            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'curr':
            if self.attributes['ndims'] == 3:
                tmp = np.reshape(self.__read_binary_data(n, 'magx'), cm)
                wy  = (np.roll(tmp, -1, axis=0) \
                     - np.roll(tmp,  1, axis=0))
                wz  = (np.roll(tmp,  1, axis=1) \
                     - np.roll(tmp, -1, axis=1))
                tmp = np.reshape(self.__read_binary_data(n, 'magy'), cm)
                wz += (np.roll(tmp, -1, axis=2) \
                     - np.roll(tmp,  1, axis=2))
                wx  = (np.roll(tmp,  1, axis=0) \
                     - np.roll(tmp, -1, axis=0))
                tmp = np.reshape(self.__read_binary_data(n, 'magz'), cm)
                wx += (np.roll(tmp, -1, axis=1) \
                     - np.roll(tmp,  1, axis=1))
                wy += (np.roll(tmp,  1, axis=2) \
                     - np.roll(tmp, -1, axis=2))
            else:
                tmp = np.reshape(self.__read_binary_data(n, 'magx'), cm)
                wz  = (np.roll(tmp,  1, axis=0) \
                     - np.roll(tmp, -1, axis=0))
                tmp = np.reshape(self.__read_binary_data(n, 'magy'), cm)
                wz += (np.roll(tmp, -1, axis=1) \
                     - np.roll(tmp,  1, axis=1))
                tmp = np.reshape(self.__read_binary_data(n, 'magz'), cm)
                wx  = (np.roll(tmp, -1, axis=0) \
                     - np.roll(tmp,  1, axis=0))
                wy  = (np.roll(tmp,  1, axis=1) \
                     - np.roll(tmp, -1, axis=1))

            dset = 0.5 * np.sqrt(wx**2 + wy**2 + wz**2)
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'dbxdx':
            p = self.attributes['ndims'] - 1
            tmp = np.reshape(self.__read_binary_data(n, 'magx'), cm)
            dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'dbxdy':
            p = self.attributes['ndims'] - 2
            tmp = np.reshape(self.__read_binary_data(n, 'magx'), cm)
            dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'dbxdz':
            p = self.attributes['ndims'] - 3
            tmp = np.reshape(self.__read_binary_data(n, 'magx'), cm)
            if p >= 0:
                dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
                dset *= 0.5
                for p in range(self.chunks[n]['dblocks']):
                    dset[...,p] /= self.cell_size[self.level[ids[p]]]
            else:
                dset = np.zeros_like(tmp)
        elif var == 'dbydx':
            p = self.attributes['ndims'] - 1
            tmp = np.reshape(self.__read_binary_data(n, 'magy'), cm)
            dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'dbydy':
            p = self.attributes['ndims'] - 2
            tmp = np.reshape(self.__read_binary_data(n, 'magy'), cm)
            dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'dbydz':
            p = self.attributes['ndims'] - 3
            tmp = np.reshape(self.__read_binary_data(n, 'magy'), cm)
            if p >= 0:
                dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
                dset *= 0.5
                for p in range(self.chunks[n]['dblocks']):
                    dset[...,p] /= self.cell_size[self.level[ids[p]]]
            else:
                dset = np.zeros_like(tmp)
        elif var == 'dbzdx':
            p = self.attributes['ndims'] - 1
            tmp = np.reshape(self.__read_binary_data(n, 'magz'), cm)
            dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'dbzdy':
            p = self.attributes['ndims'] - 2
            tmp = np.reshape(self.__read_binary_data(n, 'magz'), cm)
            dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
            dset *= 0.5
            for p in range(self.chunks[n]['dblocks']):
                dset[...,p] /= self.cell_size[self.level[ids[p]]]
        elif var == 'dbzdz':
            p = self.attributes['ndims'] - 3
            tmp = np.reshape(self.__read_binary_data(n, 'magz'), cm)
            if p >= 0:
                dset = np.roll(tmp, -1, axis=p) - np.roll(tmp,  1, axis=p)
                dset *= 0.5
                for p in range(self.chunks[n]['dblocks']):
                    dset[...,p] /= self.cell_size[self.level[ids[p]]]
            else:
                dset = np.zeros_like(tmp)
        else:
            dset = self.__read_binary_data(n, var)
            dset = np.reshape(dset, cm)

        return ids, dset


    def dataset(self, var, extent=None, maxlev=-1, shrink=1, interpolation='rebin', order=3, progress=False):

        if not var in self.variables:
            print("Dataset '{}' not available in the snapshot!".format(var))
            print("Available variables are: ", self.variables)
            return None

        dlo = np.array([self.attributes['xmin'], self.attributes['ymin']])
        dup = np.array([self.attributes['xmax'], self.attributes['ymax']])
        dln = np.array([self.attributes['xlen'], self.attributes['ylen']])
        if self.attributes['ndims'] == 3:
            dlo = np.append(dlo, self.attributes['zmin'])
            dup = np.append(dup, self.attributes['zmax'])
            dln = np.append(dln, self.attributes['zlen'])

        slo = np.array(dlo)
        sup = np.array(dup)
        if extent != None:
            if len(extent) != 2 * self.attributes['ndims']:
                print("Wrong dimension of the parameter 'extent' of the function dataset()!")
                sys.exit(1)
            slo = np.array(extent)[0::2]
            sup = np.array(extent)[1::2]
            if any(slo > dup) or any(sup < dlo) or any (slo >= sup):
                print("Wrong values in the parameter 'extent' of the function dataset()!")
                sys.exit(1)

        if isinstance(maxlev,int):
            if 1 <= maxlev <= self.attributes['maxlev']:
                shrink = 2**(self.attributes['maxlev']-maxlev)
            elif maxlev > self.attributes['maxlev']:
                sys.stdout.write("Parameter 'maxlev' should be between 1 and {}.\n".format(self.attributes['maxlev']))
                return None
        else:
            sys.stdout.write("Parameter 'maxlev' must be an integer between 1 and {}.\n".format(self.attributes['maxlev']))
            return None

        bm = np.array([ self.attributes['ncells'] ]*self.attributes['ndims'])

        if self.attributes['ndims'] == 3:
            rm = np.array([self.attributes['zblocks'], self.attributes['yblocks'], self.attributes['xblocks']])
        else:
            rm = np.array([self.attributes['yblocks'], self.attributes['xblocks']])

        dm = rm * self.attributes['ncells'] * 2**(self.ml - 1) // shrink
        ll = np.array(np.floor(dm[::-1] * (slo - dlo) / dln), dtype='int')
        uu = np.array( np.ceil(dm[::-1] * (sup - dlo) / dln), dtype='int')
        dm  = (uu - ll)[::-1]
        if var == 'velocity_vector' or var == 'magnetic_vector':
            arr = [ np.zeros(dm[:]), np.zeros(dm[:]), np.zeros(dm[:]) ]
        else:
            arr = np.zeros(dm[:])

        if progress:
            sys.stdout.write("Snapshot's path:\n  '{}'\n".format(self.path))
            m = 0

        for n in range(self.attributes['nprocs']):

            if self.chunks[n]['dblocks'] > 0:

                cm = [ self.attributes['bcells'] ]*self.attributes['ndims']
                cm.append( self.chunks[n]['dblocks'] )

                ids, dset = self.__get_dataset(n, var, cm)

                for p in range(self.chunks[n]['dblocks']):
                    ii = ids[p]
                    nl = 2**(self.ml - self.level[ii])
                    if nl <= shrink:
                        method = 'rebin'
                    else:
                        method = interpolation
                    cm = bm[0:self.attributes['ndims']] * nl // shrink
                    il = self.coord[ii][0:self.attributes['ndims']] * cm[0:self.attributes['ndims']]
                    iu = il + cm[0:self.attributes['ndims']]

                    if all(iu[:] > ll[:]) and all(il[:] < uu[:]):
                        nb = il[:] - ll[:]
                        ne = iu[:] - ll[:]
                        ib = np.maximum(nb[:], 0)
                        ie = np.minimum(ne[:], uu[:] - ll[:])
                        jb = ib[:] - nb[:]
                        je = ie[:] - ne[:] + cm[:]

                        if isinstance(dset, (list, tuple)):
                            for di in range(3):
                                if self.attributes['ndims'] == 3:
                                    arr[di][ib[2]:ie[2],ib[1]:ie[1],ib[0]:ie[0]] = interpolate(dset[di][:,:,:,p], cm, self.attributes['nghosts'], method=method, order=order)[jb[2]:je[2],jb[1]:je[1],jb[0]:je[0]]
                                else:
                                    arr[di][ib[1]:ie[1],ib[0]:ie[0]] = interpolate(dset[di][:,:,p], cm, self.attributes['nghosts'], method=method, order=order)[jb[1]:je[1],jb[0]:je[0]]
                        else:
                            if self.attributes['ndims'] == 3:
                                arr[ib[2]:ie[2],ib[1]:ie[1],ib[0]:ie[0]] = interpolate(dset[:,:,:,p], cm, self.attributes['nghosts'], method=method, order=order)[jb[2]:je[2],jb[1]:je[1],jb[0]:je[0]]
                            else:
                                arr[ib[1]:ie[1],ib[0]:ie[0]] = interpolate(dset[:,:,p], cm, self.attributes['nghosts'], method=method, order=order)[jb[1]:je[1],jb[0]:je[0]]

                    if progress:
                        mfile = 'datablocks_{:06d}.xml'.format(n)
                        m += 1
                        sys.stdout.write('\r')
                        sys.stdout.write("Reading '{}' from '{}': block {} from {}".format(var, mfile, m, self.attributes['nleafs']))
                        sys.stdout.flush()

        if (progress):
            sys.stdout.write('\n')
            sys.stdout.flush()

        return arr


    def dataset_to_vtk(self, var, label=None, compression='none', compression_level=19, progress=False):

        from .octree import OcBase, OcNode
        from .vtkio import WriteVTK
        import os

        if not var in self.variables:
            print("Dataset '{}' not available in the snapshot!".format(var))
            print("Available variables are: ", self.variables)
            return None

        if label == None:
            label = var

        if self.attributes['ndims'] < 3:
            print("Conversion to OverlappedAMR works only with 3D snapshots!")
            return None

        base = OcBase([self.attributes['xmin'], self.attributes['ymin'], self.attributes['zmin']], \
                      [self.attributes['xlen'], self.attributes['ylen'], self.attributes['zlen']], \
                      [self.attributes['xblocks'], self.attributes['yblocks'], self.attributes['zblocks']])

        ng = self.attributes['nghosts']

        if progress:
            sys.stdout.write("Snapshot's path:\n  '{}'\n".format(self.path))
            m = 0

        for n in range(self.attributes['nprocs']):

            if self.chunks[n]['dblocks'] > 0:

                cm = [ self.attributes['bcells'] ]*self.attributes['ndims']
                cm.append( self.chunks[n]['dblocks'] )

                ids, dset = self.__get_dataset(n, var, cm)

                for p in range(self.chunks[n]['dblocks']):

                    ii = self.index[ids[p]]
                    lv = self.level[ids[p]] - 1

                    center = (self.bounds[0,:,ii] + self.bounds[1,:,ii]) / 2
                    base.createNodeBranch(center, lv)
                    if isinstance(dset, (list, tuple)):
                        base.setNodeData(center, lv, (dset[0][ng:-ng,ng:-ng,ng:-ng,p], dset[1][ng:-ng,ng:-ng,ng:-ng,p], dset[2][ng:-ng,ng:-ng,ng:-ng,p]))
                    else:
                        base.setNodeData(center, lv, dset[ng:-ng,ng:-ng,ng:-ng,p])

                    if progress:
                        mfile = 'datablocks_{:06d}.xml'.format(n)
                        m += 1
                        sys.stdout.write('\r')
                        sys.stdout.write("Reading '{}' from '{}': block {} from {}".format(var, mfile, m, self.attributes['nleafs']))
                        sys.stdout.flush()

        if (progress):
            sys.stdout.write('\n')
            sys.stdout.flush()

        if progress:
            sys.stdout.write("Populating AMR structure\n")
        base.populateNodeData()

        if progress:
            sys.stdout.write("Generating OverlappingAMR VTK files\n")

        bm = np.array([ self.attributes['ncells'] ]*self.attributes['ndims'])

        if self.attributes['ndims'] == 3:
            rm = np.array([self.attributes['zblocks'], self.attributes['yblocks'], self.attributes['xblocks']])
        else:
            rm = np.array([self.attributes['yblocks'], self.attributes['xblocks']])

        ofile = "{}_{:06d}.vthb".format(var, self.attributes['isnap'])
        opath = "{}_{:06d}".format(var, self.attributes['isnap'])
        if not op.exists(opath):
            os.makedirs(opath)
        with open(ofile, 'w') as vtk:
            vtk.write('<VTKFile type="vtkOverlappingAMR" version="1.1" ' + \
                        'byte_order="LittleEndian" header_type="UInt64">\n')
            vtk.write('  <vtkOverlappingAMR ' + \
                        'origin="{} {} {}" '.format(*base.lower) + \
                        'grid_description="XYZ">\n')

            fmt = '{}_{:0' + str(len(str(self.ml))) + '}_{:0' + str(len(str(base.nodes))) + 'd}.vti'

            m = 0
            for lv in range(self.ml):

                sm = rm[::-1] * 2**lv
                bw = base.size / sm
                cw = bw / self.attributes['ncells']
                vtk.write('    <Block level="{}"'.format(lv) + \
                                ' spacing="{} {} {}">\n'.format(*cw))

                no = 0
                for item in base.getNodesFromLevel(lv):
                    lo = np.array(item.index) * bm
                    up = lo + bm - 1
                    ll = np.stack((lo,up)).T.flatten()
                    if item.hasData:
                        vfile = op.join(opath, fmt.format(var, lv, no))
                        WriteVTK(vfile, label, item.data, \
                            origin = (item.lower[0], item.lower[1], item.lower[2]), \
                            spacing = (cw[0], cw[1], cw[2]), \
                            compression=compression, compression_level=compression_level)
                        vtk.write('      <DataSet index="{}"'.format(no) + \
                                    ' amr_box = "{} {} {} {} {} {}"'.format(*ll) + \
                                        ' file = "{}"></DataSet>\n'.format(vfile))
                        no += 1
                    else:
                        vtk.write('      <DataSet index="{}"'.format(no) + \
                                    ' amr_box = "{} {} {} {} {} {}"'.format(*ll) + \
                                        '></DataSet>\n')
                    m += 1

                    if progress:
                        sys.stdout.write('\r')
                        sys.stdout.write("Storing AMR block {} from {}".format(m, base.nodes))
                        sys.stdout.flush()

                vtk.write('    </Block>\n')
            vtk.write('  </vtkOverlappingAMR>\n')
            vtk.write('</VTKFile>')

        if (progress):
            sys.stdout.write('\n')
            sys.stdout.flush()
