"""
================================================================================

  This file is part of the AMUN source code, a program to perform
  Newtonian or relativistic magnetohydrodynamical simulations on uniform or
  adaptive mesh.

  Copyright (C) 2018-2021 Grzegorz Kowal <grzegorz@amuncode.org>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

================================================================================

 module: AMUN

  Python module with subroutines to read AMUN code HDF5 files.

  The only requirements for this package are:

     - numpy
     - scipy (optional, for data interpolation)

--------------------------------------------------------------------------------
"""
import numpy as np
try:
    from scipy.ndimage import zoom
    from scipy.interpolate import splrep, splev, interp1d, pchip_interpolate
    scipy_available = True
except ImportError:
    scipy_available = False


def rebin(a, newshape):
  '''
      Subroutine changes the size of the input array to to new shape,
      by copying cells or averaging them.
  '''
  assert len(a.shape) == len(newshape)

  m = a.ndim - 1
  if (a.shape[m] > newshape[m]):
    if a.ndim == 3:
      nn = [newshape[0], int(a.shape[0] / newshape[0]),
            newshape[1], int(a.shape[1] / newshape[1]),
            newshape[2], int(a.shape[2] / newshape[2])]
      return a.reshape(nn).mean(5).mean(3).mean(1)
    else:
      nn = [newshape[0], int(a.shape[0] / newshape[0]),
            newshape[1], int(a.shape[1] / newshape[1])]
      return a.reshape(nn).mean(3).mean(1)
  else:
    for n in range(a.ndim):
      a = np.repeat(a, int(newshape[n] / a.shape[n]), axis=n)
    return(a)


def interpolate(a, newshape, nghosts, method='rebin', order=3):
  '''
      Subroutine rescales the block by interpolating its values.
  '''
  if (method == 'rebin' or not scipy_available):

    # calculate the indices in order to remove the ghost zones
    #
    if a.ndim == 3:
      ib = nghosts
      jb = nghosts
      kb = nghosts
      ie = a.shape[2] - nghosts
      je = a.shape[1] - nghosts
      ke = a.shape[0] - nghosts
      if (a.shape[0] == 1):
        kb = 0
        ke = 1

      return rebin(a[kb:ke,jb:je,ib:ie], newshape)
    else:
      ib = nghosts
      jb = nghosts
      ie = a.shape[1] - nghosts
      je = a.shape[0] - nghosts

      return rebin(a[jb:je,ib:ie], newshape)

  elif (method == 'zoom'):

    fc = int(newshape[1] / (a.shape[1] - 2 * nghosts))
    ib, ie = fc * nghosts, - fc * nghosts
    jb, je = fc * nghosts, - fc * nghosts
    if a.ndim == 3:
      kb, ke = fc * nghosts, - fc * nghosts
      if (a.shape[0] == 1):
        kb = 0
        ke = 1

      return zoom(a, fc, order=order, grid_mode=True, mode='nearest')[kb:ke,jb:je,ib:ie]
    else:
      return zoom(a, fc, order=order, grid_mode=True, mode='nearest')[jb:je,ib:ie]

  elif (method == 'monotonic' or method == 'pchip'):

    dims = np.arange(a.ndim)

    q = a

    for n in dims:

      d2 = np.roll(q,-1, axis=0) + np.roll(q, 1, axis=0) - 2.0 * q
      q  = q - d2 / 24.0

      d  = np.array(q.shape)

      xo = (np.arange(0.5, a.shape[n]) - nghosts) / (a.shape[n] - 2 * nghosts)
      xn =  np.arange(0.5, newshape[n]) / newshape[n]

      u = q.reshape([d[0], int(q.size / d[0])])
      f = np.zeros([newshape[n], int(q.size / d[0])])
      for i in range(int(q.size / d[0])):
        f[:,i] = pchip_interpolate(xo, u[:,i], xn)

      d[0] = newshape[n]
      f = f.reshape(d)

      q = f.transpose(np.roll(dims, -1))

    return q

  elif (method == 'spline'):

    dims = np.arange(a.ndim)

    q = a

    for n in dims:

      d2 = np.roll(q,-1, axis=0) + np.roll(q, 1, axis=0) - 2.0 * q
      q  = q - d2 / 24.0

      d  = np.array(q.shape)

      xo = (np.arange(0.5, a.shape[n]) - nghosts) / (a.shape[n] - 2 * nghosts)
      xn =  np.arange(0.5, newshape[n]) / newshape[n]

      u = q.reshape([d[0], int(q.size / d[0])])
      f = np.zeros([newshape[n], int(q.size / d[0])])
      for i in range(int(q.size / d[0])):
        t = splrep(xo, u[:,i], k=5, s=0.0)
        f[:,i] = splev(xn, t)

      d[0] = newshape[n]
      f = f.reshape(d)

      q = f.transpose(np.roll(dims, -1))

    return q

  else:

    dims = np.arange(a.ndim)

    q = a

    for n in dims:

      d2 = np.roll(q,-1, axis=0) + np.roll(q, 1, axis=0) - 2.0 * q
      q  = q - d2 / 24.0

      d  = np.array(q.shape)

      xo = (np.arange(0.5, a.shape[n]) - nghosts) / (a.shape[n] - 2 * nghosts)
      xn =  np.arange(0.5, newshape[n]) / newshape[n]

      u = q.reshape([d[0], int(q.size / d[0])])
      f = np.zeros([newshape[n], int(q.size / d[0])])
      for i in range(int(q.size / d[0])):
        t = interp1d(xo, u[:,i], kind=method)
        f[:,i] = t(xn)

      d[0] = newshape[n]
      f = f.reshape(d)

      q = f.transpose(np.roll(dims, -1))

    return q
