"""
================================================================================

  This file is part of the AMUN source code, a program to perform
  Newtonian or relativistic magnetohydrodynamical simulations on uniform or
  adaptive mesh.

  Copyright (C) 2018-2021 Grzegorz Kowal <grzegorz@amuncode.org>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

================================================================================

  submodule: vtkio.py

  This submodule provides a function to store a dataset as VTK image file.

  The only requirements for this package are:

     - numpy
     - lz4 (for LZ4 compression)

--------------------------------------------------------------------------------
"""
import base64, numpy, struct, zlib, lz4.block, lzma

def WriteVTK(vtkfile, vname, data, origin = (0.0, 0.0, 0.0), \
            spacing = (1.0, 1.0, 1.0), points = False, encode = True, \
            compression = "none", block_size = 32768, lz4mode = 'fast', \
            compression_level = 6, verbose = False):

    # Separate cases when the input is a vector field and scalar field.
    #
    if isinstance(data, (list, tuple)):

        # Check if all list or tuple components are arrays.
        #
        allarrays = True
        for d in data:
            allarrays = allarrays & isinstance(d, (numpy.ndarray))

        # Quit if not all elements are arrays.
        #
        if not allarrays:

            print('All input data components must be arrays!')
            return

        # Check if the components have the same dimensions
        #
        sameshapes = True
        for d in data:
            sameshapes = sameshapes & (data[0].shape == d.shape)

        # Quit if elements have different dimensions.
        #
        if not sameshapes:
            print('All input data component must have the same dimensions!')
            return

        # Define the type of input data.
        #
        dtype = 'vector'

        # Get the number of vector components.
        #
        nc = len(data)

        # Get the number of dimansions
        #
        nd = data[0].ndim

        # Get the variable dimensions.
        #
        dm = data[0].shape

    elif isinstance(data, (numpy.ndarray)):

        # Define the type of input data.
        #
        dtype = 'scalar'

        # Get the number of vector components.
        #
        nc = 1

        # Get the number of dimansions
        #
        nd = data.ndim

        # Get the variable dimensions.
        #
        dm = data.shape

    else:
        print('Unknown type of the input data!')
        return

    # Create the VTK output file and open it in the binary mode.
    #
    with open(vtkfile, 'wb') as vt:

        # Initiate offset and array size.
        #
        offset = 0

        # Prepare strong for dimensions.
        #
        sdims  = '"'
        if points:
            for i in range(nd - 1, 0, -1):
                sdims +='%d %d ' % (0, dm[i] - 1)
            sdims += '%d %d" ' % (0, dm[0] - 1)
        else:
            for i in range(nd - 1, 0, -1):
                sdims +='%d %d ' % (0, dm[i])
            sdims += '%d %d" ' % (0, dm[0])

        # Write the VTK header and data description.
        #
        string  = '<?xml version="1.0"?>\n<VTKFile type="ImageData" version="1.0" byte_order="LittleEndian" header_type="UInt64"'
        if compression == 'zlib':
            string  += ' compressor="vtkZLibDataCompressor"'
        elif compression == 'lzma':
            string  += ' compressor="vtkLZMADataCompressor"'
        elif compression == 'lz4':
            string  += ' compressor="vtkLZ4DataCompressor"'
        string += '>\n  <ImageData WholeExtent=%s' % sdims
        string += 'Origin="%e %e %e" ' % origin
        string += 'Spacing="%e %e %e">\n' % spacing
        string += '    <Piece Extent=%s>\n' % sdims
        if points:
            string += '      <PointData %ss="%s">\n' % (dtype, vname)
        else:
            string += '      <CellData %ss="%s">\n' % (dtype, vname)
        if nc == 1:
            dmin = data.min()
            dmax = data.max()
        else:
            dd   = numpy.zeros(data[0].shape)
            for dc in data:
                dd += dc**2
            dmin = numpy.sqrt(dd.min())
            dmax = numpy.sqrt(dd.max())
        string += '        <DataArray '
        if nc > 1:
            string += 'NumberOfComponents="{:d}" '.format(nc)
        string += 'type="Float32" Name="{}" RangeMin="{:e}" RangeMax="{:e}" format="appended" offset="{}">\n'.format(vname, dmin, dmax, offset)
        string += "        </DataArray>\n"
        if points:
            string += '      </PointData>\n'
        else:
            string += '      </CellData>\n'
        string += '    </Piece>\n'
        string += '  </ImageData>\n'
        vt.write(string.encode())

        # Append variable data.
        #
        if encode:
            string  = '  <AppendedData encoding="base64">\n'
        else:
            string  = '  <AppendedData encoding="raw">\n'
        string += '   _'
        vt.write(string.encode())

        # Convert the input data to Float32. In the case of vector data concatenate the components first.
        #
        if (dtype == 'vector'):
            qt = numpy.zeros([ dm[0], dm[1], dm[2], nc ], dtype = numpy.float32)
            for n, d in enumerate(data[:]):
                qt[:,:,:,n] = numpy.float32(d)

        else:
            qt = numpy.float32(data)

        barr = qt.tobytes()

        # Compress if desired.
        #
        if compression != 'none':

            lz4compression = 10 - compression_level

            if len(barr) < block_size:

                if compression == 'zlib':
                    carr = zlib.compress(barr, compression_level)
                elif compression == 'lz4':
                    carr = lz4.block.compress(barr, mode = lz4mode, acceleration = lz4compression, compression = compression_level, store_size = False)
                elif compression == 'lzma':
                    carr = lzma.compress(barr)

                # Prepare the number of blocks, the size of the last partial block,
                # and the size of the compressed data.
                #
                head = struct.pack("QQQQ", 1, len(barr), 0, len(carr))

            else:

                nblocks = len(barr) // block_size
                rsize   = len(barr) % block_size
                if verbose:
                    print("Number of blocks:\t {}\nBlock size:\t\t {}\nRemaining size:\t\t {}".format(nblocks, block_size, rsize))

                head = struct.pack("QQQ", nblocks, block_size, rsize)

                carr = bytearray(b'')
                ib = 0
                if compression == 'zlib':
                    cctx = zlib
                    for i in range(nblocks):
                        ie = min(len(barr), ib + block_size)
                        cblk = cctx.compress(barr[ib:ie], compression_level)
                        head += struct.pack("Q", len(cblk))
                        carr += cblk
                        ib = ie

                elif compression == 'lzma':
                    cctx = lzma
                    for i in range(nblocks):
                        ie = min(len(barr), ib + block_size)
                        cblk = cctx.compress(barr[ib:ie], preset = compression_level)
                        head += struct.pack("Q", len(cblk))
                        carr += cblk
                        ib = ie

                elif compression == 'lz4':
                    cctx = lz4.block
                    for i in range(nblocks):
                        ie = min(len(barr), ib + block_size)
                        cblk = cctx.compress(barr[ib:ie], mode = lz4mode, acceleration = lz4compression, compression = compression_level, store_size = False)
                        head += struct.pack("Q", len(cblk))
                        carr += cblk
                        ib = ie

            if encode:
                vt.write(base64.b64encode(head)+base64.b64encode(carr))
            else:
                vt.write(head+carr)

        else:

            head = struct.pack("Q", len(barr))

            if encode:
                vt.write(base64.b64encode(head+barr))
            else:
                vt.write(head+barr)

        # Close the appended data section.
        #
        string  = '\n  </AppendedData>\n'
        string += '</VTKFile>\n'
        vt.write(string.encode())
